#include "optimize.h"

#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <vector>

#include <levmar.h>

namespace Bilinear
{

static FILE *fr = 0;



double func_poly_1(double *p){return 1;}

double func_poly_x(double *p){return p[0];}
double func_poly_y(double *p){return p[1];}
double func_poly_z(double *p){return p[2];}

double func_poly_xx(double *p){return p[0]*p[0];}
double func_poly_yy(double *p){return p[1]*p[1];}
double func_poly_zz(double *p){return p[2]*p[2];}
double func_poly_yz(double *p){return p[1]*p[2];}
double func_poly_xz(double *p){return p[0]*p[2];}
double func_poly_xy(double *p){return p[0]*p[1];}
typedef double (*poly_func_type)(double *p);
poly_func_type poly_func[10] = 
{func_poly_1, 

func_poly_x, func_poly_y, func_poly_z,

func_poly_xx, func_poly_yy, func_poly_zz,
func_poly_yz, func_poly_xz, func_poly_xy
};


//#define CONSTRAINED_UV_PENALTY


typedef double PolyCoeffs[bilinear_npoly];

struct SetErrorBi
{
	double error, error_max;
	int index;
	BoundedArray<int, bilinear_nfuncs> seti; // indexes basisindexbi
	BoundedArray<double, bilinear_nfuncs*bilinear_npoly*3> params;
	int subdomain[3];

	void calc_constant_coeff()
	{
		params.s = bilinear_nfuncs*bilinear_npoly*3; // set to be capacity... always
		PolyCoeffs *c = (PolyCoeffs*)&params[2*bilinear_nfuncs*bilinear_npoly];
		
		for (int i = 0; i < bilinear_npoly; i++)
			c[bilinear_nfuncs-1][i] = 0;
		c[bilinear_nfuncs-1][0] = 1;

		for (int j = 0; j < bilinear_nfuncs-1; j++)
		{
			for (int i = 0; i < bilinear_npoly; i++)
				c[bilinear_nfuncs-1][i] -= c[j][i];
		}
	}
	
	bool operator<(const SetErrorBi &a) const
	{
		return error < a.error;
	}
};

struct PatchBi
{
	Array3D<Sample> samples;
	SetErrorBi best_set;
	Array<Affinity> affinity;
};

static PatchBi patch;
static Array2D<double> bb;

struct BilinearOptData
{
	BasisIndexBi *bi;
	int nbi;
	bool constant_precision;
};

double eval_poly(PolyCoeffs &c, double *x)
{
	if (bilinear_npoly == 1)
	{
		return c[0];
	}
	else if (bilinear_npoly == 4)
	{
		return c[0] + c[1]*x[0] + c[2]*x[1] + c[3]*x[2];
	}
	else if (bilinear_npoly == 10)
	{
		return c[0] + c[1]*x[0] + c[2]*x[1] + c[3]*x[2]
		 + c[4]*x[0]*x[0] + c[5]*x[1]*x[1] + c[6]*x[2]*x[2]
		 + c[7]*x[1]*x[2] + c[8]*x[0]*x[1] + c[9]*x[0]*x[1];
	}
	else
	{
		double r = 0;
		for (int i = 0; i < bilinear_npoly; i++)
			r += c[i]*poly_func[i](x);
		return r;
	}
}

double eval_poly_der(PolyCoeffs &c, double *x, int d)
{
	if (bilinear_npoly == 1)
	{
		return 1;
	}
	else if (bilinear_npoly == 4)
	{
		if (d == 0)
			return 1;
		else
			return x[d-1];
	}
	else
	{
		return poly_func[d](x);
	}
}

struct SquareIndex
{
	int i, s; // index, size
};

Array<SquareIndex> square_indices;
Array<CompletedSquare> squares;
double sum_constant;


void calc_error_max(double &error_sum, double &error_max, PatchBi &patch, BoundedArray<int, bilinear_nfuncs*4> &seti, BoundedArray<double, bilinear_nfuncs*bilinear_npoly*3> &params)
{
	static Array2D<double> A;
	static Array<double> B;
	static Array<double> X;
	double C;

	error_sum = 0;
	error_max = 0;
	
	const int nbi = bilinear_nfuncs;
	PolyCoeffs *u = (PolyCoeffs*)&params[0];
	PolyCoeffs *v = (PolyCoeffs*)&params[nbi*bilinear_npoly];
	PolyCoeffs *c = (PolyCoeffs*)&params[2*nbi*bilinear_npoly];

	// sum errors for all points in the domain. Lots of little systems of equations.
	X.resize(bilinear_nfuncs*4);
	A.resize(X.s,X.s);
	B.resize(X.s);

	for (int i = 0; i < X.s; i++)
	{
		for (int j = 0; j < X.s; j++)
			A(i,j) = bb(seti[i],seti[j]);
	}

	for (int k = 0; k < patch.samples.data_size; k++)
	{
		Sample &sample = patch.samples.data[k];

		// build A, B, C
		C = sample.hh;
		for (int i = 0; i < X.s; i++)
			B[i] = sample.hb[seti[i]];

		// build X
		for (int ib = 0; ib < nbi; ib++)
		{
			double uu = eval_poly(u[ib], sample.t);
			double vv = eval_poly(v[ib], sample.t);
			double cc;

			if (ib < nbi - 1)
			{
				cc = eval_poly(c[ib], sample.t);
			}
			else
			{
				cc = 1;
				for (int j = 0; j < nbi - 1; j++)
					cc -= eval_poly(c[j], sample.t);
			}

			X[4*ib + 0] = cc * (1 - uu) * (1 - vv);
			X[4*ib + 1] = cc * uu * (1 - vv);
			X[4*ib + 2] = cc * (1 - uu) * vv;
			X[4*ib + 3] = cc * uu * vv;
		}

		// eval error
		double error = C;
		for (int i = 0; i < X.s; i++)
		{
			double Ax = 0;
			for (int j = 0; j < X.s; j++)
				Ax += X[j] * A(i,j);
			error += X[i] * (Ax - 2*B[i]);
		}

		error_sum += error;
		error_max = max(error_max, error);
	}

	error_sum /= patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);
}

void build_error_data(void *datavoidptr)
{
	squares.s = 0;
	square_indices.s = 0;

	BilinearOptData *dataptr = (BilinearOptData *)datavoidptr;
	BasisIndexBi *bi = dataptr->bi;
	int nbi = dataptr->nbi;
	bool constant_precision = dataptr->constant_precision;

	Array2D<double> A;
	Array<double> B;

	A.resize(4*nbi,4*nbi);
	B.resize(4*nbi);

	// A doesn't depend on the sample position, so build it out of the loop
	for (int ib = 0; ib < nbi; ib++)
	{
		for (int i = 0; i < 4; i++)
		{
			for (int jb = 0; jb < nbi; jb++)
			{
				for (int j = 0; j < 4; j++)
					A(4*ib + i, 4*jb + j) = bb(bi[ib].idx[i], bi[jb].idx[j]);
			}
		}
	}
	
	Array<CompletedSquare> squares_tmp;
	SquareIndex idx;
	idx.i = 0;
	sum_constant = 0;
	for (int is = 0; is < patch.samples.data_size; is++)
	{
		Sample &s = patch.samples.data[is];
		
		// sub matrices
		for (int ib = 0; ib < nbi; ib++)
		{
			for (int i = 0; i < 4; i++)
				B[4*ib + i] = -2*s.hb[bi[ib].idx[i]];
		}
		
		//mat_print(A, "A.txt");
		//vec_print(B, "B.txt");

		// complete squares
		squares_tmp.s = 0;
		double const_val = complete_squares(squares_tmp, A, B, s.hh);
		sum_constant += const_val;

		//printf("const of square is %f -> %f\n", s.hh, const_val);

		// store
		for (int i = 0; i < squares_tmp.s; i++)
		{
			CompletedSquare &cs = squares_tmp[i];
			squares.push_back(cs);

			//printf("%f: ", cs.constant);
			//for (int j = 0; j < cs.linear.s; j++)
			//	printf(" %f", cs.linear[j]);
			//printf("\n");
		}
		//printf("\n");

		idx.s = squares_tmp.s;
		square_indices.push_back(idx);
		idx.i += idx.s;
	}

	sum_constant /= (double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);

	//printf("DONE BUILDING SQUARES - constant = %f\n", sum_constant);
}

void error(double *params, double *vals, int mmm, int nnn, void *datavoidptr)
{
	BilinearOptData *dataptr = (BilinearOptData *)datavoidptr;
	BasisIndexBi *bi = dataptr->bi;
	int nbi = dataptr->nbi;
	bool constant_precision = dataptr->constant_precision;
	
	PolyCoeffs *u = (PolyCoeffs*)&params[0];
	PolyCoeffs *v = (PolyCoeffs*)&params[nbi*bilinear_npoly];
	PolyCoeffs *c = (PolyCoeffs*)&params[2*nbi*bilinear_npoly];

	Array<double> X;
	X.resize(4*nbi);

	for (int is = 0; is < patch.samples.data_size; is++)
	{
		Sample &s = patch.samples.data[is];

		//printf("s.t = %f, %f, %f; s.h = %f, %f, %f\n", s.t[0], s.t[1], s.t[2], s.h[0], s.h[1], s.h[2]);
		
		// X from parameters
		for (int ib = 0; ib < nbi; ib++)
		{
			double uu = eval_poly(u[ib], s.t);
			double vv = eval_poly(v[ib], s.t);
			double cc;

			if (!constant_precision || ib < nbi - 1)
			{
				cc = eval_poly(c[ib], s.t);
			}
			else
			{
				cc = 1;
				for (int j = 0; j < nbi - 1; j++)
					cc -= eval_poly(c[j], s.t);
			}

			X[4*ib + 0] = cc * (1 - uu) * (1 - vv);
			X[4*ib + 1] = cc * uu * (1 - vv);
			X[4*ib + 2] = cc * (1 - uu) * vv;
			X[4*ib + 3] = cc * uu * vv;
		}

		//vec_print(X, "X_bi.txt");
		//exit(0);

		// eval
		SquareIndex idx = square_indices[is];
		for (int ss = idx.i; ss < idx.i + idx.s; ss++)
		{
			CompletedSquare &s = squares[ss];

			vals[ss] = s.constant;

			for (int i = 0; i < s.linear.s; i++)
			{
				vals[ss] += s.linear[i] * X[i];
			}

			assert(_finite(vals[ss]));
		}
	}

	for (int i = 0; i < squares.s; i++)
		vals[i] /= sqrt((double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z));

#ifdef CONSTRAINED_UV_PENALTY
	// add the penalty term if the UV parameters go out of range
	vals[squares.s] = 0;
	for (int z = 0; z < 2; z++)
	{
		for (int y = 0; y < 2; y++)
		{
			for (int x = 0; x < 2; x++)
			{
				for (int i = 0; i < bilinear_nfuncs*2; i++)
				{
					double val = params[i*bilinear_npoly + 0] +
								x*params[i*bilinear_npoly + 1] +
								y*params[i*bilinear_npoly + 2] +
								z*params[i*bilinear_npoly + 3];
					if (val < 0)
						vals[squares.s] -= val;
					else if (val > 1)
						vals[squares.s] += val - 1;

					
				}
			}
		}
	}
	vals[squares.s] *= 1000;

	vals[squares.s] += sqrt(sum_constant);
#endif
	
//#if 1
//	// calc and print error
//	double e = 0;
//	for (int i = 0; i < squares.s + 1; i++)
//		e += vals[i]*vals[i];
//	printf("ERROR = %f        \n", e);
//
//	exit(0);
//	return;
//#endif
}




void calculate_affinities(Array<Affinity> &affinity, Array<BasisIndexBi> &basisindexbi)
{
	for (int i = 0; i < basisindexbi.s; i++)
	{
		Affinity &a = affinity[i];
		BilinearOptData bod;
		
		bod.nbi = 1;
		bod.bi = &basisindexbi[i];
		bod.constant_precision = false;

		Array<double> p;
		p.resize(3*bilinear_npoly);
		for (int ii = 0; ii < p.s; ii++)
			p[ii] = 0;

		double info[LM_INFO_SZ];
		build_error_data(&bod);
		dlevmar_dif(error, &p[0], NULL, p.s, squares.s, bilinear_iter_max, NULL, info, NULL, NULL, &bod);
		//dlevmar_der(error, jacobian, &p[0], NULL, p.s, patch.samples.data_size, bilinear_iter_max, NULL, info, NULL, NULL, &bod);

#if 0
		for (int j = 0; j < p.s; j++)
			printf("p[%d] = %f\n", j, p[j]);
		printf("\n");

		for (int j = 0; j < LM_INFO_SZ; j++)
			printf("info[%d] = %f\n", j, info[j]);
		printf("\n");
#endif

		a.affinity = info[1];
		a.index = i;

		printf("affinity[%d] = %f   iters = %d      \r", a.index, -a.affinity, (int)info[5]);
	}

	sort(affinity.begin(), affinity.end());
/*
	for (int i = 0; i < affinity.s; i++)
		printf("affinity[%d] = %f\n", affinity[i].index, -affinity[i].affinity);*/
}


int get_sets_tree(SetErrorBi &best_set,
				   BoundedArray<int, bilinear_nfuncs> &seti,
				   Array<BasisIndexBi> &basisindexbi,
				   Array<Affinity> &order,
				   int mini,
				   int nleaves,
				   PatchBi &patch)
{
	if (nleaves >= bilinear_maxleaves)
		return nleaves;

	if (seti.s == bilinear_nfuncs)
	{
		printf("Trying new basis set\n");
		SetErrorBi sete;
		SetErrorBi sete_restart;
		sete_restart.error = 1e30;

		//double t0 = get_time();

		for (int rand_sample_iter = 0; rand_sample_iter < 1; rand_sample_iter++)
		{
			// resize
			if (rand_sample_iter == 0)
			{
				int para_num = (bilinear_nfuncs * 3 - 1) * bilinear_npoly;
				sete.params.resize(para_num);
				for (int i = 0; i < sete.params.s; i++)
					sete.params[i] = 0;
			}
			else
			{
				// use stored values for the restart plus some randomization
				for (int i = 0; i < sete.params.s; i++)
					sete.params[i] = sete_restart.params[i] + .5*frand0();
			}

			//** calculate
			double info[LM_INFO_SZ]; // levmar info stored here

			BilinearOptData bod;

			BasisIndexBi bib[bilinear_nfuncs];
			for (int i = 0; i < bilinear_nfuncs; i++)
				bib[i] = basisindexbi[seti[bilinear_nfuncs-1-i]]; // reversed order because opposite traversal from mathematica
			bod.nbi = bilinear_nfuncs;
			bod.bi = bib;
			bod.constant_precision = true;


#if 0
			// DEBUG -- set parameters and basis functions to reproduce an example
			bib[0].idx[0] = 11 - 1;
			bib[0].idx[1] = 16 - 1;
			bib[0].idx[2] = 12 - 1;
			bib[0].idx[3] = 17 - 1;

			bib[1].idx[0] = 17 - 1;
			bib[1].idx[1] = 22 - 1;
			bib[1].idx[2] = 18 - 1;
			bib[1].idx[3] = 23 - 1;

			sete.params[0] = 0.461424;
			sete.params[1] = 0.354362;
			sete.params[2] = -0.328033;
			sete.params[3] = 0.184102;
			sete.params[4] = 0.22642;
			sete.params[5] = 0.31988;
			sete.params[6] = -0.009614;
			sete.params[7] = -0.216783;
			sete.params[8] = 0.461424;
			sete.params[9] = -0.328033;
			sete.params[10] = 0.354362;
			sete.params[11] = 0.184102;
			sete.params[12] = 0.22642;
			sete.params[13] = -0.009614;
			sete.params[14] = 0.31988;
			sete.params[15] = -0.216783;
			sete.params[16] = 0.979766;
			sete.params[17] = -0.454796;
			sete.params[18] = -0.454796;
			sete.params[19] = -0.338632;
#endif

			// build data for optimization
			build_error_data(&bod);

#ifdef CONSTRAINED_UV_PENALTY
			dlevmar_dif(error, &sete.params[0], NULL, sete.params.s, squares.s + 1, bilinear_iter_max, NULL, info, NULL, NULL, &bod);
			sete.error = info[1];
#else
			dlevmar_dif(error, &sete.params[0], NULL, sete.params.s, squares.s    , bilinear_iter_max, NULL, info, NULL, NULL, &bod);
			sete.error = info[1] + sum_constant;
#endif

			printf("iters = %f   error = %f\n", info[5], sete.error);


			//** store
			if (nleaves % 1 == 0)
				printf(" nleaves = %d\r", nleaves);

			//printf(" set = {%d, %d}\n", seti[0], seti[1]);
			
			if (sete.error < sete_restart.error)
			{
				printf("new best error current set %f at leaf %d                 \n", sete.error, nleaves);
				sete_restart.error = sete.error;
				sete_restart.params = sete.params;
			}

			if (sete.error < best_set.error)
			{
				printf("+++ new overall best error %f at leaf %d +++                \n", sete.error, nleaves);

				for (int i = 0; i < bilinear_nfuncs; i++)
					best_set.seti[i] = seti[bilinear_nfuncs-1-i]; // reversed order because opposite traversal from mathematica

				best_set.error = sete.error;
				best_set.params = sete.params;
				best_set.index = nleaves;

				best_set.calc_constant_coeff();

				// save the data to a log
				fprintf(fr, "\n\n========== Iteration %d ===========\n\n", nleaves);
				fprintf(fr, "error = %f\n", sete.error);

				// check max error
				double error_sum, error_max;
				BoundedArray<int, 4*bilinear_nfuncs> bib_flat;
				for (int i = 0; i < bilinear_nfuncs; i++)
					for (int j = 0; j < 4; j++)
						bib_flat[i*4 + j] = bib[i].idx[j];
				calc_error_max(error_sum, error_max, patch, bib_flat, sete.params);
				best_set.error_max = error_max;
				fprintf(fr, "error_max = %f\n", sete.error_max);

				// write funcs
				for (int i = 0; i < bilinear_nfuncs; i++)
					fprintf(fr, "bib[%d] = {%d, %d, %d, %d}\n", i, bib[i].idx[0]+1, bib[i].idx[1]+1, bib[i].idx[2]+1, bib[i].idx[3]+1);
				fprintf(fr, "\n");

#if defined(CONSTRAINED_UV_PENALTY)
				// check corners for satisfying constraints
				fprintf(fr, "constraints check (all should be zero)\n");
				double constraint_sum = 0;
				for (int z = 0; z < 2; z++)
				{
					for (int y = 0; y < 2; y++)
					{
						for (int x = 0; x < 2; x++)
						{
							for (int i = 0; i < bilinear_nfuncs*2; i++)
							{
								double val = sete.params[i*bilinear_npoly + 0] +
									x*sete.params[i*bilinear_npoly + 1] +
									y*sete.params[i*bilinear_npoly + 2] +
									z*sete.params[i*bilinear_npoly + 3];
								if (val < 0)
									;
								else if (val > 1)
									val--;
								else
									val = 0;
								fprintf(fr, "%f ", val);
								constraint_sum += val*val;
							}
						}
					}
				}
				fprintf(fr, "\nconstraint_sum^2 = %f\n\n", constraint_sum);
#endif

				fprintf(fr, "params = {%f", sete.params[0]);
				for (int i = 1; i < sete.params.s; i++)
					fprintf(fr, ", %f", sete.params[i]);
				fprintf(fr, "}\n\n");

				for (int i = 0; i < LM_INFO_SZ; i++)
					fprintf(fr, "info[%d] = %f\n", i, info[i]);
				fprintf(fr, "\n");
				fflush(fr);

				//exit(0);
			}

		}

		nleaves++;
		
		//double t1 = get_time();
		//printf("Time to calculate iteration = %f\n", t1 - t0);

		return nleaves;
	}

	for (int i = basisindexbi.s-1; i >= mini; i--)
	{
		seti.push_back(order[i].index);
		nleaves = get_sets_tree(best_set, seti, basisindexbi, order, i+1, nleaves, patch);
		seti.pop_back();
	}

	return nleaves;
}


void get_sets(PatchBi &patch, Array<BasisIndexBi> &basisindexbi)
{
	patch.best_set.error = 1e30;

	BoundedArray<int, bilinear_nfuncs> seti;
	get_sets_tree(patch.best_set, seti, basisindexbi, patch.affinity, 0, 0, patch);
}

}
using namespace Bilinear;

void calc_bilinear_fetches()
{
	//========== choose basis functions
	Array<BasisParam> basisparams;
	Array<BasisIndexBi> basisindexbi;

	choose_basis_functions(basisparams);
	choose_bilinear_basis_indices(basisindexbi, basisparams);

	// select which patches to evaluate. -1 means all patches in that dimension.
	int inpatch_i = -1, inpatch_j = -1, inpatch_k = -1;
	//int inpatch_i = 0, inpatch_j = 0, inpatch_k = 0;

	Array<SetErrorBi> best_sets;
	best_sets.reserve(6);

	char outname[256];

	// build full linear systems
	for (int pk = 0; pk < npatches_z; pk++)
	{
		if (inpatch_k != -1 && inpatch_k != pk)
			continue;

		for (int pj = 0; pj < npatches_xy; pj++)
		{
			if (inpatch_j != -1 && inpatch_j != pj)
				continue;

			for (int pi = 0; pi < npatches_xy; pi++)
			{
				if (inpatch_i != -1 && inpatch_i != pi)
					continue;

				if (pj > pi)
					continue;
				
				sprintf(outname, "results%s_patch_bi_%d_%d_%d.txt", max_mip_level == 1 ? "_low" : "", pi, pj, pk);
				fr = fopen(outname, "wb");
				printf("opening file %s for writing\n", outname);

				printf("starting now for patch %d %d %d\n", pi, pj, pk);
				fprintf(fr, "starting now for patch %d %d %d\n", pi, pj, pk);
				double t0 = get_time();

				// patch offsets and scales
				double poi = pi / (double)(2 * npatches_xy);
				double psi = 1 / (double)(2 * npatches_xy);

				double poj = pj / (double)(2 * npatches_xy);
				double psj = 1 / (double)(2 * npatches_xy);

				//double pok = 1 + pk / (double)(npatches_z);
				double pok = pk / (double)(npatches_z);
				double psk = 1 / (double)(npatches_z);
				
				//========== build overlap matrices
				build_overlaps(patch.samples, bb, basisparams, poi, psi, poj, psj, pok, psk, nsamples);

				double t1 = get_time();
				printf("build overlaps = %f\n", t1 - t0);
				fprintf(fr, "build overlaps = %f\n", t1 - t0);

				//========== affinities
				// init to be in mathematica order
				patch.affinity.resize(basisindexbi.s);
				for (int i = 0; i < basisindexbi.s; i++)
				{
					patch.affinity[i].index = basisindexbi.s - 1 - i;
					patch.affinity[i].affinity = 0;
				}

				// calculate better order
				//calculate_affinities(patch.affinity, basisindexbi);
				
				//========== find best set
				get_sets(patch, basisindexbi);

				// store best set
				patch.best_set.subdomain[0] = pi;
				patch.best_set.subdomain[1] = pj;
				patch.best_set.subdomain[2] = pk;

				best_sets.push_back(patch.best_set);

				fclose(fr);
			}
		}
	}

	
	// print total error
	printf("nfuncs = %d, npoly = %d, npatches_xy = %d (%d), npatches_z = %d\n", bilinear_nfuncs, bilinear_npoly, npatches_xy, npatches_xy*2, npatches_z);
	double total_error = 0;
	double maximal_error = 0;
	for (int i = 0; i < best_sets.s; i++)
	{
		if (best_sets[i].subdomain[0] == best_sets[i].subdomain[1])
			total_error += 4*best_sets[i].error; // only 4x on diagonal, because we cover the whole square rather than just a triangle
		else
			total_error += 8*best_sets[i].error;

		maximal_error = max(maximal_error, best_sets[i].error_max);
	}
	printf("The total error over all patches is %f\n", total_error);
	printf("The maximal error over all patches is %f\n", maximal_error);

	//** output C++ arrays of best sets in symmetric form 6
	sprintf(outname, "results%s_all_bi.txt", max_mip_level == 1 ? "_low" : "");
	fr = fopen(outname, "wb");
	printf("opening file %s for writing\n", outname);
	fprintf(fr, "The total error over all patches is %f\n\n", total_error);
	fprintf(fr, "The maximal error over all patches is %f\n", maximal_error);

	// write indices
	fprintf(fr, "static int indices[%d][%d][3] = {\n", best_sets.s, bilinear_nfuncs);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetErrorBi &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < bilinear_nfuncs; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			BasisParam bp = basisparams[basisindexbi[s.seti[i]].idx[0]];
			fprintf(fr, "{%d,%d,%d}", bp.i, bp.j, bp.k);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	// write coefficients
	fprintf(fr, "static double params[%d][%d] = {\n", best_sets.s, best_sets[0].params.s);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetErrorBi &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < s.params.s; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			fprintf(fr, "%16.15f", s.params[i]);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	fclose(fr);
}
