#include "optimize.h"

#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <vector>

#include <levmar.h>

namespace BilinearHW
{

static FILE *fr = 0;



double func_poly_1(double *p){return 1;}

double func_poly_x(double *p){return p[0];}
double func_poly_y(double *p){return p[1];}
double func_poly_z(double *p){return p[2];}

double func_poly_xx(double *p){return p[0]*p[0];}
double func_poly_yy(double *p){return p[1]*p[1];}
double func_poly_zz(double *p){return p[2]*p[2];}
double func_poly_yz(double *p){return p[1]*p[2];}
double func_poly_xz(double *p){return p[0]*p[2];}
double func_poly_xy(double *p){return p[0]*p[1];}
typedef double (*poly_func_type)(double *p);
poly_func_type poly_func[10] = 
{func_poly_1, 

func_poly_x, func_poly_y, func_poly_z,

func_poly_xx, func_poly_yy, func_poly_zz,
func_poly_yz, func_poly_xz, func_poly_xy
};

//#define CONSTRAINED_UV
//#define HARDWARE_BILINEAR


typedef double PolyCoeffs[bilinear_npoly];

struct SetErrorBiHW
{
	double error, error_max;
	int index;
	BoundedArray<int, bilinear_nfuncs> seti; // indexes basisindexbi
	BoundedArray<double, bilinear_nfuncs*bilinear_npoly*3> params;
	int subdomain[3];

	void calc_constant_coeff()
	{
		params.s = bilinear_nfuncs*bilinear_npoly*3; // set to be capacity... always
		PolyCoeffs *c = (PolyCoeffs*)&params[2*bilinear_nfuncs*bilinear_npoly];
		
		for (int i = 0; i < bilinear_npoly; i++)
			c[bilinear_nfuncs-1][i] = 0;
		c[bilinear_nfuncs-1][0] = 1;

		for (int j = 0; j < bilinear_nfuncs-1; j++)
		{
			for (int i = 0; i < bilinear_npoly; i++)
				c[bilinear_nfuncs-1][i] -= c[j][i];
		}
	}
	
	bool operator<(const SetErrorBiHW &a) const
	{
		return error < a.error;
	}
};

struct PatchBiHW
{
	Array3D<Sample> samples;
	SetErrorBiHW best_set;
	Array<Affinity> affinity;
};

static PatchBiHW patch;
static Array2D<double> bb;

struct BilinearHWOptData
{
	int *mip_levels;
	int nbi;
	bool constant_precision;
};

double eval_poly(PolyCoeffs &c, double *x)
{
	if (bilinear_npoly == 1)
	{
		return c[0];
	}
	else if (bilinear_npoly == 4)
	{
		return c[0] + c[1]*x[0] + c[2]*x[1] + c[3]*x[2];
	}
	else if (bilinear_npoly == 10)
	{
		return c[0] + c[1]*x[0] + c[2]*x[1] + c[3]*x[2]
		 + c[4]*x[0]*x[0] + c[5]*x[1]*x[1] + c[6]*x[2]*x[2]
		 + c[7]*x[1]*x[2] + c[8]*x[0]*x[1] + c[9]*x[0]*x[1];
	}
	else
	{
		double r = 0;
		for (int i = 0; i < bilinear_npoly; i++)
			r += c[i]*poly_func[i](x);
		return r;
	}
}

double eval_poly_der(PolyCoeffs &c, double *x, int d)
{
	if (bilinear_npoly == 1)
	{
		return 1;
	}
	else if (bilinear_npoly == 4)
	{
		if (d == 0)
			return 1;
		else
			return x[d-1];
	}
	else
	{
		return poly_func[d](x);
	}
}



void calc_error_max(double &error_sum, double &error_max, PatchBiHW &patch, BoundedArray<int, bilinear_nfuncs> &mip_levels, BoundedArray<double, bilinear_nfuncs*bilinear_npoly*3> &params)
{
	static Array2D<double> A;
	static Array<double> B;
	static Array<double> X;
	double C;

	error_sum = 0;
	error_max = 0;
	
	const int nbi = bilinear_nfuncs;
	PolyCoeffs *u = (PolyCoeffs*)&params[0];
	PolyCoeffs *v = (PolyCoeffs*)&params[nbi*bilinear_npoly];
	PolyCoeffs *c = (PolyCoeffs*)&params[2*nbi*bilinear_npoly];

	// sum errors for all points in the domain. Lots of little systems of equations.
	X.resize(bb.s[0]);
	BasisIndexBi bi[bilinear_nfuncs];

	for (int k = 0; k < patch.samples.data_size; k++)
	{
		Sample &s = patch.samples.data[k];

		// X from parameters
		X = 0;

		for (int ib = 0; ib < nbi; ib++)
		{
			// relative offset of u,v coordinates
			double uo = eval_poly(u[ib], s.t);
			double vo = eval_poly(v[ib], s.t);

			// position of sample in different levels (absolute in frame)
			int mip = mip_levels[ib];

			double ua = -.5 - grid_basis_indices[mip].minidx + uo;
			double va = -.5 - grid_basis_indices[mip].minidx + vo;

			// check that ua, va are valid
			double uu, vv;
			int ui, vi;

			if (ua < 0)
			{
				uu = 0;
				ui = 0;
			}
			else if (ua >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				ui = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				uu = 1;
			}
			else
			{
				ui = floor(ua);
				uu = ua - ui;
			}

			if (va < 0)
			{
				vv = 0;
				vi = 0;
			}
			else if (va >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				vi = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				vv = 1;
			}
			else
			{
				vi = floor(va);
				vv = va - vi;
			}

			// use (ui, vi) to choose the basis function indices (bi)
			bi[ib].idx[0] = grid_basis_indices[mip].grid(ui,vi);
			bi[ib].idx[1] = grid_basis_indices[mip].grid(ui+1,vi);
			bi[ib].idx[2] = grid_basis_indices[mip].grid(ui,vi+1);
			bi[ib].idx[3] = grid_basis_indices[mip].grid(ui+1,vi+1);

			// coefficient
			double cc;

			if (ib < nbi - 1)
			{
				cc = eval_poly(c[ib], s.t);
			}
			else
			{
				cc = 1;
				for (int j = 0; j < nbi - 1; j++)
					cc -= eval_poly(c[j], s.t);
			}

			// bilinear coefficients of texel quad

			X[bi[ib].idx[0]] += cc * (1 - uu) * (1 - vv);
			X[bi[ib].idx[1]] += cc * uu * (1 - vv);
			X[bi[ib].idx[2]] += cc * (1 - uu) * vv;
			X[bi[ib].idx[3]] += cc * uu * vv;
		}
		
		// calc error
		double error = s.hh;
		for (int i = 0; i < X.s; i++)
		{
			error -= 2 * s.hb[i] * X[i];

			for (int j = 0; j < X.s; j++)
				error += bb(i,j) * X[i] * X[j];
		}

		error_sum += error;
		error_max = max(error_max, error);
	}

	error_sum /= patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);
}



#if 0 // Chose between using local basis functions and using all basis functions

// build A and B using only the local basis functions. this can cause the error vector to change discontinuously, which is probably bad for the optimization
void error_hw_sample(double *params, double *vals, int mmm, int nnn, void *datavoidptr)
{
	BilinearHWOptData *dataptr = (BilinearHWOptData *)datavoidptr;
	int *mip_levels = dataptr->mip_levels;
	int nbi = dataptr->nbi;
	bool constant_precision = dataptr->constant_precision;

	PolyCoeffs *u = (PolyCoeffs*)&params[0];
	PolyCoeffs *v = (PolyCoeffs*)&params[nbi*bilinear_npoly];
	PolyCoeffs *c = (PolyCoeffs*)&params[2*nbi*bilinear_npoly];

	Array<double> X;
	X.resize(4*nbi);
	
	BasisIndexBi bi[bilinear_nfuncs];
	double sum_constant = 0;
	double boundary_penalty = 0;
	int val_index = 0;
	for (int is = 0; is < patch.samples.data_size; is++)
	{
		Sample &s = patch.samples.data[is];

		// X from parameters
		for (int ib = 0; ib < nbi; ib++)
		{
			// relative offset of u,v coordinates
			double uo = eval_poly(u[ib], s.t);
			double vo = eval_poly(v[ib], s.t);

			// position of sample in different levels (absolute in frame)
			int mip = mip_levels[ib];
			double ua = s.h[0] * ((npatches_xy*2) << mip) - .5 - grid_basis_indices[mip].minidx + uo;
			double va = s.h[1] * ((npatches_xy*2) << mip) - .5 - grid_basis_indices[mip].minidx + vo;

			// check that ua, va are valid
			double uu, vv;
			int ui, vi;

			if (ua < 0)
			{
				boundary_penalty -= ua;
				uu = 0;
				ui = 0;
			}
			else if (ua >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				boundary_penalty += ua - (grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx);
				ui = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				uu = 1;
			}
			else
			{
				ui = floor(ua);
				uu = ua - ui;
			}

			if (va < 0)
			{
				boundary_penalty -= va;
				vv = 0;
				vi = 0;
			}
			else if (va >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				boundary_penalty += va - (grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx);
				vi = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				vv = 1;
			}
			else
			{
				vi = floor(va);
				vv = va - vi;
			}

			// use (ui, vi) to choose the basis function indices (bi)
			bi[ib].idx[0] = grid_basis_indices[mip].grid(ui,vi);
			bi[ib].idx[1] = grid_basis_indices[mip].grid(ui+1,vi);
			bi[ib].idx[2] = grid_basis_indices[mip].grid(ui,vi+1);
			bi[ib].idx[3] = grid_basis_indices[mip].grid(ui+1,vi+1);

			// coefficient
			double cc;

			if (!constant_precision || ib < nbi - 1)
			{
				cc = eval_poly(c[ib], s.t);
			}
			else
			{
				cc = 1;
				for (int j = 0; j < nbi - 1; j++)
					cc -= eval_poly(c[j], s.t);
			}

			// bilinear coefficients of texel quad
			X[4*ib + 0] = cc * (1 - uu) * (1 - vv);
			X[4*ib + 1] = cc * uu * (1 - vv);
			X[4*ib + 2] = cc * (1 - uu) * vv;
			X[4*ib + 3] = cc * uu * vv;
		}

		//*** eval (need to complete squares each time)
		Array2D<double> A;
		Array<double> B;

		A.resize(4*nbi,4*nbi);
		B.resize(4*nbi);

		// both A and B depend on the sample position now, so build it for each sample
		for (int ib = 0; ib < nbi; ib++)
		{
			for (int i = 0; i < 4; i++)
			{
				for (int jb = 0; jb < nbi; jb++)
				{
					for (int j = 0; j < 4; j++)
						A(4*ib + i, 4*jb + j) = bb(bi[ib].idx[i], bi[jb].idx[j]);
				}
			}
		}
		
		for (int ib = 0; ib < nbi; ib++)
		{
			for (int i = 0; i < 4; i++)
				B[4*ib + i] = -2*s.hb[bi[ib].idx[i]];
		}
		
		//mat_print(A, "A.txt");
		//vec_print(B, "B.txt");


		// complete squares
		Array<CompletedSquare> squares;
		double const_val = complete_squares(squares, A, B, s.hh);
		sum_constant += const_val;

		// multiply completed squares by X
		for (int ss = 0; ss < squares.s; ss++, val_index++)
		{
			CompletedSquare &s = squares[ss];

			vals[val_index] = s.constant;

			for (int i = 0; i < s.linear.s; i++)
			{
				vals[val_index] += s.linear[i] * X[i];
			}

			assert(_finite(vals[val_index]));
		}
	}

	// make any remaining vals 0
	for (; val_index < 8*patch.samples.data_size; val_index++)
		vals[val_index] = 0;

	for (int i = 0; i < 8*patch.samples.data_size; i++)
		vals[i] /= sqrt((double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z));
	sum_constant /= (double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);
	
	vals[8*patch.samples.data_size] = sqrt(sum_constant) + 1000*boundary_penalty;

#if 0
	// calc and print error
	double e = 0;
	for (int i = 0; i < 8*patch.samples.data_size + 1; i++)
		e += vals[i]*vals[i];
	printf("ERROR = %f\n", e);
	return;
#endif
}

#else

struct SquareIndex
{
	int i, s; // index, size
};

Array<SquareIndex> square_indices;
Array<CompletedSquare> squares;
double sum_constant;
int levmar_iter = 0;

void build_error_data(void *datavoidptr)
{
	squares.s = 0;
	square_indices.s = 0;

	BilinearHWOptData *dataptr = (BilinearHWOptData *)datavoidptr;
	int *mip_levels = dataptr->mip_levels;
	int nbi = dataptr->nbi;
	bool constant_precision = dataptr->constant_precision;

	Array<double> b2;

	Array<CompletedSquare> squares_tmp;
	SquareIndex idx;
	idx.i = 0;
	sum_constant = 0;
	for (int is = 0; is < patch.samples.data_size; is++)
	{
		Sample &s = patch.samples.data[is];
		
		//mat_print(bb, "A.txt");
		//vec_print(s.hb, "B.txt");
		//exit(0);

		// complete squares
		b2 = s.hb;
		for (int i = 0; i < b2.s; i++)
			b2[i] *= -2;

		squares_tmp.s = 0;
		double const_val = complete_squares(squares_tmp, bb, b2, s.hh);
		sum_constant += const_val;

		//printf("const of square is %f -> %f\n", s.hh, const_val);

		// store
		for (int i = 0; i < squares_tmp.s; i++)
		{
			CompletedSquare &cs = squares_tmp[i];
			squares.push_back(cs);

			//printf("%f: ", cs.constant);
			//for (int j = 0; j < cs.linear.s; j++)
			//	printf(" %f", cs.linear[j]);
			//printf("\n");
		}
		//printf("\n");

		idx.s = squares_tmp.s;
		square_indices.push_back(idx);
		idx.i += idx.s;

		////***** check that completed squares are correct
		//int n = bb.s[0];
		//Array2D<double> a;
		//a.resize(n,n);
		//a = 0;
		//Array<double> b;
		//b.resize(n);
		//b = 0;
		//double c = 0;

		//for (int i = 0; i < squares_tmp.s; i++)
		//{
		//	CompletedSquare &cs = squares_tmp[i];

		//	c += cs.constant * cs.constant;
		//	for (int r = 0; r < n; r++)
		//	{
		//		b[r] += cs.linear[r] * cs.constant;
		//		for (int c = 0; c < n; c++)
		//		{
		//			a(r,c) += cs.linear[r] * cs.linear[c];
		//		}
		//	}
		//}

		//mat_print(a, "Acs.txt");
		//vec_print(b, "Bcs.txt");
		//exit(0);
	}

	sum_constant /= (double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);

	//printf("DONE BUILDING SQUARES - constant = %f\n", sum_constant);
}

double geomean(double a, double b)
{
	return sqrt(a*a + b*b);
}

// use a much larger system of equations for A and B that uses all of the basis functions. Most of the x vector will be zero, but the error vector will be continuous.
void error_hw_sample(double *params, double *vals, int mmm, int nnn, void *datavoidptr)
{
	BilinearHWOptData *dataptr = (BilinearHWOptData *)datavoidptr;
	int *mip_levels = dataptr->mip_levels;
	int nbi = dataptr->nbi;
	bool constant_precision = dataptr->constant_precision;

	PolyCoeffs *u = (PolyCoeffs*)&params[0];
	PolyCoeffs *v = (PolyCoeffs*)&params[nbi*bilinear_npoly];
	PolyCoeffs *c = (PolyCoeffs*)&params[2*nbi*bilinear_npoly];

	Array<double> X;
	X.resize(bb.s[0]);
	
	BasisIndexBi bi[bilinear_nfuncs];
	double boundary_penalty = 0;
	int val_index = 0;
	//double error2 = 0; // debug var for direct calculation
	for (int is = 0; is < patch.samples.data_size; is++)
	{
		Sample &s = patch.samples.data[is];

		// X from parameters
		X = 0;

		for (int ib = 0; ib < nbi; ib++)
		{
			// relative offset of u,v coordinates
			double uo = eval_poly(u[ib], s.t);
			double vo = eval_poly(v[ib], s.t);

			// position of sample in different levels (absolute in frame)
			int mip = mip_levels[ib];

			double ua = -.5 - grid_basis_indices[mip].minidx + uo;
			double va = -.5 - grid_basis_indices[mip].minidx + vo;

			// check that ua, va are valid
			double uu, vv;
			int ui, vi;

			if (ua < 0)
			{
				boundary_penalty -= ua;
				uu = 0;
				ui = 0;
			}
			else if (ua >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				boundary_penalty += ua - (grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx);
				ui = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				uu = 1;
			}
			else
			{
				ui = floor(ua);
				uu = ua - ui;
			}

			if (va < 0)
			{
				boundary_penalty -= va;
				vv = 0;
				vi = 0;
			}
			else if (va >= grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx)
			{
				boundary_penalty += va - (grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx);
				vi = grid_basis_indices[mip].maxidx - grid_basis_indices[mip].minidx - 1;
				vv = 1;
			}
			else
			{
				vi = floor(va);
				vv = va - vi;
			}

			// use (ui, vi) to choose the basis function indices (bi)
			bi[ib].idx[0] = grid_basis_indices[mip].grid(ui,vi);
			bi[ib].idx[1] = grid_basis_indices[mip].grid(ui+1,vi);
			bi[ib].idx[2] = grid_basis_indices[mip].grid(ui,vi+1);
			bi[ib].idx[3] = grid_basis_indices[mip].grid(ui+1,vi+1);

			// coefficient
			double cc;

			if (!constant_precision || ib < nbi - 1)
			{
				cc = eval_poly(c[ib], s.t);
			}
			else
			{
				cc = 1;
				for (int j = 0; j < nbi - 1; j++)
					cc -= eval_poly(c[j], s.t);
			}

			// bilinear coefficients of texel quad

			X[bi[ib].idx[0]] += cc * (1 - uu) * (1 - vv);
			X[bi[ib].idx[1]] += cc * uu * (1 - vv);
			X[bi[ib].idx[2]] += cc * (1 - uu) * vv;
			X[bi[ib].idx[3]] += cc * uu * vv;
		}
		

		//vec_print(X, "X_bihw.txt");
		//exit(0);

		//*** eval
		SquareIndex idx = square_indices[is];
		//printf("idx = %d, size = %d\n", idx, square_indices[is].s);
		for (int ss = idx.i; ss < idx.i + idx.s; ss++)
		{
			CompletedSquare &s = squares[ss];

			vals[ss] = s.constant;

			for (int i = 0; i < s.linear.s; i++)
			{
				vals[ss] += s.linear[i] * X[i];
			}

			assert(_finite(vals[ss]));
		}

		//// direct calc (debug)
		//error2 += s.hh;
		//for (int i = 0; i < X.s; i++)
		//{
		//	error2 -= 2 * s.hb[i] * X[i];

		//	for (int j = 0; j < X.s; j++)
		//		error2 += bb(i,j) * X[i] * X[j];
		//}
	}

	for (int i = 0; i < squares.s; i++)
		vals[i] /= sqrt((double)patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z));
	
	vals[squares.s] = sqrt(sum_constant) + 1000*boundary_penalty;

#if 1
	// calc and print error
	double e = 0;
	for (int i = 0; i < squares.s + 1; i++)
		e += vals[i]*vals[i];
	printf("ERROR = %f  iter = %d        \r", e, levmar_iter);
	levmar_iter++;

	//error2 /= patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);
	//exit(0);
	//return;
#endif
}

#endif



#define LINEAR_INITIAL
#ifdef LINEAR_INITIAL
// LINEAR INITIAL GUESS SEEDS
static int indices_guess[][2][3] = {
	{{-1,-1,1},{0,0,1}},
	{{-1,0,1},{0,-1,1}},
	{{-1,-1,1},{0,0,1}},
	{{-1,-1,0},{-1,-1,1}},
	{{0,-1,1},{1,0,2}},
	{{0,0,1},{1,1,2}}
};

static double params_guess[][24] = {
	{0.388000154952278,0.261095159324691,-0.041113439305281,0.164278568791913,0.266243883660315,0.394427680008516,-0.054972321328907,-0.211271562297314,0.388000154949017,-0.041113439306042,0.261095159330546,0.164278568792342,0.266243883737881,-0.054972321399706,0.394427679973804,-0.211271562338174,0.825569231099933,-0.164045015060200,-0.164045015051735,0.279854851042283,0.174430768900067,0.164045015060200,0.164045015051735,-0.279854851042283},
	{0.000000000000000,0.000000000000000,0.000000000000000,0.000000000000000,0.317944973692554,0.241482792198133,0.018460464681539,-0.131502961855246,0.000000000000000,0.000000000000000,0.000000000000000,0.000000000000000,0.437362047009008,0.008285803095218,0.240590595626637,0.110066980951919,0.190727303009453,-0.134111274419508,0.044555131439078,-0.139360211835419,0.809272696990547,0.134111274419508,-0.044555131439078,0.139360211835419},
	{0.517933356477268,0.321089579929378,-0.271929510425802,0.160953369532935,0.402851552124436,0.234412237088360,-0.059069442735105,-0.155248682551179,0.517933356422583,-0.271929510410708,0.321089580035806,0.160953369782026,0.402851552114579,-0.059069442728629,0.234412237092891,-0.155248682547227,0.540008938051265,-0.213044000193063,-0.213044000195342,-0.281404584336323,0.459991061948735,0.213044000193063,0.213044000195342,0.281404584336323},
	{0.317746094519650,0.007726715009744,-0.021649682407589,0.354330060224244,0.447714013330732,0.398567889753355,0.017912676289198,0.081687852916111,0.317746094619479,-0.021649682705756,0.007726714685905,0.354330060413273,0.447714013337754,0.017912676277932,0.398567889736880,0.081687852922119,-0.461552969556927,0.322518899306480,0.322518899299365,-0.355947656526968,1.461552969556927,-0.322518899306480,-0.322518899299365,0.355947656526968},
	{0.101129691557106,0.237015900292026,0.021281735165960,-0.064234492311765,0.379088270382929,0.309046780960661,0.015826059920396,-0.122659370956603,0.394879276829317,0.086041437807766,0.322485713611307,0.172265793951048,0.307434162056292,-0.091006904364438,0.253283664553625,-0.216427257691849,0.978796032981226,-0.102931650966455,-0.121564334902376,-0.102289182049419,0.021203967018774,0.102931650966455,0.121564334902376,0.102289182049419},
	{0.129395013904662,0.298009555978122,-0.016284106327490,-0.113110907570355,0.491019406650816,0.282025810954199,-0.124073923469815,-0.211628303213385,0.129395013906991,-0.016284106333254,0.298009555980520,-0.113110907573737,0.491019406654929,-0.124073923471791,0.282025810951181,-0.211628303216373,1.089638572771813,-0.230927755252642,-0.230927755252718,-0.161631545063640,-0.089638572771813,0.230927755252642,0.230927755252718,0.161631545063640}
};
#else
// QUADRATIC INITIAL GUESS from linear
static int mip_levels_guess[][2] = 

static double params_guess[][24] = 
#endif

// seed with an initial guess that was the best constrained optimization
int get_sets_tree(SetErrorBiHW &best_set,
				   BoundedArray<int, bilinear_nfuncs> &seti,
				   int nleaves,
				   PatchBiHW &patch, int iter)
{
	SetErrorBiHW sete;
	printf("\n\n");

#ifdef LINEAR_INITIAL
	seti.s = 2;
	seti[0] = indices_guess[iter][0][2];
	seti[1] = indices_guess[iter][1][2];
#else
	seti.s = bilinear_nfuncs;
	for (int i = 0; i < bilinear_nfuncs; i++)
		seti[i] = mip_levels_guess[iter][i];
#endif

	for (int rand_sample_iter = 0; rand_sample_iter < 20; rand_sample_iter++)
	{
		printf("rand_sample_iter %d\n", rand_sample_iter);
		//continue;
		assert(bilinear_nfuncs == 2);

		//double t0 = get_time();
#ifdef LINEAR_INITIAL
		// set initial guess
		sete.params.resize((bilinear_nfuncs * 3 - 1) * bilinear_npoly);
		for (int i = 0; i < sete.params.s; i++)
			sete.params[i] = params_guess[iter][i];
		
		// RELATIVE TO 0,0 -- BETTER, SIMPLER OPTION?
		// u1
		sete.params[0] += indices_guess[iter][0][0] + .5;
		
		// u2
		sete.params[bilinear_npoly] += indices_guess[iter][1][0] + .5;

		// v1
		sete.params[2*bilinear_npoly] += indices_guess[iter][0][1] + .5;

		// v2
		sete.params[3*bilinear_npoly] += indices_guess[iter][1][1] + .5;
#else
		// set initial guess to zero
		sete.params.resize((bilinear_nfuncs * 3 - 1) * bilinear_npoly);
		for (int i = 0; i < sete.params.s; i++)
			sete.params[i] = 0;
		
		// fill in nonzero parts from linear solve
		for (int i = 0; i < bilinear_nfuncs * 3 - 1; i++)
		{
			// 4 polynmial coefficients in linear
			for (int j = 0; j < 4; j++)
				sete.params[i*bilinear_npoly + j] = params_guess[iter][i*4 + j];
		}
#endif

		// add some noise to start position (leave first iter alone so that we get at least as good results)
		if (rand_sample_iter > 0)
		{
			for (int i = 0; i < sete.params.s; i++)
			{
				double v = frand0() * .5;
				//printf("v = %f\n", v);
				sete.params[i] += v;
			}
		}

		//** calculate
		double info[LM_INFO_SZ]; // levmar info stored here

		BilinearHWOptData bod;
		bod.nbi = bilinear_nfuncs;
		bod.mip_levels = seti.data;
		bod.constant_precision = true;

		build_error_data(&bod);
		levmar_iter = 0;
		dlevmar_dif(error_hw_sample, &sete.params[0], NULL, sete.params.s, squares.s + 1, bilinear_iter_max, NULL, info, NULL, NULL, &bod);

		sete.error = info[1];
		printf("iters = %f   error = %f   mip_levels = {%d, %d}\n", info[5], sete.error, seti[0], seti[1]);


		//** store
		if (nleaves % 1 == 0)
			printf(" nleaves = %d\r", nleaves);

		//printf(" set = {%d, %d}\n", seti[0], seti[1]);


		if (sete.error < best_set.error)
		{
			printf("new best error %f at leaf %d                 \n", sete.error, nleaves);

			best_set.seti = seti;
			best_set.error = sete.error;
			best_set.params = sete.params;
			best_set.index = nleaves;

			best_set.calc_constant_coeff();
			
			// calc max error
			double error_sum = 0, error_max = 0;
			calc_error_max(error_sum, error_max, patch, seti, sete.params);
			best_set.error_max = error_max;
			fprintf(fr, "error_max = %f\n", sete.error_max);

#if 1
			fprintf(fr, "error = %f\n", sete.error);
			fprintf(fr, "iteration = %d\n\n", nleaves);

			fprintf(fr, "mip_levels =");
			for (int i = 0; i < bilinear_nfuncs; i++)
				fprintf(fr, " %d", seti[i]);
			fprintf(fr, "\n\n");

			for (int i = 0; i < sete.params.s; i++)
				fprintf(fr, "p[%d] = %f\n", i, sete.params[i]);
			fprintf(fr, "\n");

			for (int i = 0; i < LM_INFO_SZ; i++)
				fprintf(fr, "info[%d] = %f\n", i, info[i]);
			fprintf(fr, "\n");
			fflush(fr);
#endif

			//print_set(best_set, basisindexbi);
		}

		//double t1 = get_time();
		//printf("Time to calculate iteration = %f\n", t1 - t0);
	}

	return 0;
}


void get_sets(PatchBiHW &patch)
{
	static int iter = 0;
	patch.best_set.error = 1e30;

	// need a nested number of loops equal to bilinear_nfuncs (recusion i guess)
	BoundedArray<int, bilinear_nfuncs> seti;
	get_sets_tree(patch.best_set, seti, 0, patch, iter);

	iter++;
}


}
using namespace BilinearHW;





void calc_bilinear_hw_fetches()
{
	//========== choose basis functions
	Array<BasisParam> basisparams;
	choose_hw_basis_functions(basisparams);

	// select which patches to evaluate. -1 means all patches in that dimension.
	int inpatch_i = -1, inpatch_j = -1, inpatch_k = -1;
	//int inpatch_i = 1, inpatch_j = 0, inpatch_k = 0;

	Array<SetErrorBiHW> best_sets;
	best_sets.reserve(6);

	char outname[256];

	// build full linear systems
	for (int pk = 0; pk < npatches_z; pk++)
	{
		if (inpatch_k != -1 && inpatch_k != pk)
			continue;

		for (int pj = 0; pj < npatches_xy; pj++)
		{
			if (inpatch_j != -1 && inpatch_j != pj)
				continue;

			for (int pi = 0; pi < npatches_xy; pi++)
			{
				if (inpatch_i != -1 && inpatch_i != pi)
					continue;

				if (pj > pi)
					continue;
				
				sprintf(outname, "results%s_patch_bi_%d_%d_%d.txt", max_mip_level == 1 ? "_low" : "", pi, pj, pk);
				fr = fopen(outname, "wb");
				printf("opening file %s for writing\n", outname);

				printf("starting now for patch %d %d %d\n", pi, pj, pk);
				fprintf(fr, "starting now for patch %d %d %d\n", pi, pj, pk);
				double t0 = get_time();

				// patch offsets and scales
				double poi = pi / (double)(2 * npatches_xy);
				double psi = 1 / (double)(2 * npatches_xy);

				double poj = pj / (double)(2 * npatches_xy);
				double psj = 1 / (double)(2 * npatches_xy);

				//double pok = 1 + pk / (double)(npatches_z);
				double pok = pk / (double)(npatches_z);
				double psk = 1 / (double)(npatches_z);
				
				//========== build overlap matrices
				build_overlaps(patch.samples, bb, basisparams, poi, psi, poj, psj, pok, psk, nsamples);

				double t1 = get_time();
				printf("build overlaps = %f\n", t1 - t0);
				fprintf(fr, "build overlaps = %f\n", t1 - t0);
				
				//========== find best set
				get_sets(patch);

				// store best set
				patch.best_set.subdomain[0] = pi;
				patch.best_set.subdomain[1] = pj;
				patch.best_set.subdomain[2] = pk;

				best_sets.push_back(patch.best_set);

				fclose(fr);
			}
		}
	}

	
	// print total error
	printf("nfuncs = %d, npoly = %d, npatches_xy = %d (%d), npatches_z = %d\n", bilinear_nfuncs, bilinear_npoly, npatches_xy, npatches_xy*2, npatches_z);
	double total_error = 0;
	double maximal_error = 0;
	for (int i = 0; i < best_sets.s; i++)
	{
		if (best_sets[i].subdomain[0] == best_sets[i].subdomain[1])
			total_error += 4*best_sets[i].error; // only 4x on diagonal, because we cover the whole square rather than just a triangle
		else
			total_error += 8*best_sets[i].error;

		maximal_error = max(maximal_error, best_sets[i].error_max);
	}
	printf("The total error over all patches is %f\n", total_error);
	printf("The maximal error over all patches is %f\n", maximal_error);

	//** output C++ arrays of best sets in symmetric form 6
	sprintf(outname, "results%s_all_bi.txt", max_mip_level == 1 ? "_low" : "");
	fr = fopen(outname, "wb");
	printf("opening file %s for writing\n", outname);
	fprintf(fr, "The total error over all patches is %f\n\n", total_error);
	fprintf(fr, "The maximal error over all patches is %f\n", maximal_error);

	// write indices
	fprintf(fr, "static int mip_levels[%d][%d] = {\n", best_sets.s, bilinear_nfuncs);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetErrorBiHW &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < bilinear_nfuncs; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			fprintf(fr, "%d", s.seti[i]);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	// write coefficients
	fprintf(fr, "static double params[%d][%d] = {\n", best_sets.s, best_sets[0].params.s);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetErrorBiHW &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < s.params.s; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			fprintf(fr, "%16.15f", s.params[i]);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	fclose(fr);
}
