#include "optimize.h"

#include <stdio.h>
#include <math.h>

#include <algorithm>
#include <set>

#include <clapack.h>

using namespace std;


namespace Single
{

int dposv_joe(char uplo, integer n, integer nrhs, doublereal *a, integer lda, doublereal *b, integer ldb)
{
	integer info;
	dposv_(&uplo, &n, &nrhs, a, &lda, b, &ldb, &info);
	return info;
}






double func_poly_1(double *p){return 1;}

double func_poly_x(double *p){return p[0];}
double func_poly_y(double *p){return p[1];}
double func_poly_z(double *p){return p[2];}

double func_poly_xx(double *p){return p[0]*p[0];}
double func_poly_yy(double *p){return p[1]*p[1];}
double func_poly_zz(double *p){return p[2]*p[2];}
double func_poly_yz(double *p){return p[1]*p[2];}
double func_poly_xz(double *p){return p[0]*p[2];}
double func_poly_xy(double *p){return p[0]*p[1];}
typedef double (*poly_func_type)(double *p);
poly_func_type poly_func[10] = 
{func_poly_1, 

func_poly_x, func_poly_y, func_poly_z,

func_poly_xx, func_poly_yy, func_poly_zz,
func_poly_yz, func_poly_xz, func_poly_xy
};

const int single_npoly = 4; // 1, 4, 10
const int single_nfuncs = 8;
const int single_maxleaves = 1*1000*1000;

struct SetError
{
	double error, error_max;
	int index;
	BoundedArray<int, single_nfuncs> seti;
	BoundedArray<double, single_nfuncs*single_npoly> coeffs;
	BoundedArray<BasisParam, single_nfuncs> basis_idx;
	int subdomain[3];
	
	bool operator<(const SetError &a) const
	{
		return error < a.error;
	}
};


struct Patch
{
	Array3D<Sample> samples;
	SetError best_set;
	Array<Affinity> affinity;
};

static Array2D<double> bb;

static FILE *fr = 0;

void print_set(SetError &best_set, Array<BasisParam> &basis)
{
	// print the sets that won
	fprintf(fr, "error %f, error_max %f, index %d ->\n", best_set.error, best_set.error_max, best_set.index);

	fprintf(fr, " i = {");
	for (int j = 0; j < single_nfuncs; j++)
	{
		if (j != 0)
			fprintf(fr, ",");
		fprintf(fr, "%d", best_set.seti[j]);
	}
	fprintf(fr, "}\n");

	fprintf(fr, " k = {");
	for (int j = 0; j < single_nfuncs; j++)
	{
		BasisParam &b = basis[best_set.seti[j]];
		if (j != 0)
			fprintf(fr, ",");
		fprintf(fr, "{%d,%d,%d}", b.i, b.j, b.k);
	}
	fprintf(fr, "}\n");

	fprintf(fr, " c = {");
	BoundedArray<double, single_nfuncs*single_npoly> &c = best_set.coeffs;
	for (int j = 0; j < c.s; j++)
	{
		if (j != 0)
			fprintf(fr, ",");
		fprintf(fr, "%f", c[j]);
		if (j % 5 == 4)
			fprintf(fr, "\n ");
	}
	fprintf(fr, "}\n");
	fflush(fr);
}

Array2D<Array2D<double> > Apart;
Array<Array<double> > Bpart;
double Cpart;

void build_parts(Patch &patch, Array<BasisParam> &basis)
{
	// reserve
	Apart.resize(basis.s, basis.s);
	Bpart.resize(basis.s);

	// alloc and zero
	for (int i = 0; i < basis.s; i++)
	{
		Bpart[i].resize(single_npoly);
		Bpart[i] = 0;

		for (int j = 0; j < basis.s; j++)
		{
			Apart(i,j).resize(single_npoly, single_npoly);
			Apart(i,j) = 0;
		}
	}

	// build
	Cpart = 0;
	for (int k = 0; k < patch.samples.data_size; k++)
	{
		Sample &sample = patch.samples.data[k];
		Cpart += sample.hh;
	}

	for (int r = 0; r < basis.s; r++)
	{
		for (int s = 0; s < single_npoly; s++)
		{
			for (int k = 0; k < patch.samples.data_size; k++)
			{
				Sample &sample = patch.samples.data[k];
				Bpart[r][s] += sample.hb[r] * poly_func[s](sample.t);
			}
			for (int i = 0; i < basis.s; i++)
			{
				Array2D<double> &Apart_ri = Apart(r,i);
				double bbir = bb(i, r);
				for (int j = 0; j < single_npoly; j++)
				{
					double temp = 0;
					for (int k = 0; k < patch.samples.data_size; k++)
					{
						Sample &sample = patch.samples.data[k];
						temp += bbir * poly_func[j](sample.t) * poly_func[s](sample.t);
					}
					Apart_ri(s,j) = temp;
				}
			}
		}
	}
}

void calc_error_max(double &error_sum, double &error_max, Patch &patch, BoundedArray<int, single_nfuncs> &seti, BoundedArray<double, single_nfuncs*single_npoly> &X_in)
{
	static Array2D<double> A;
	static Array<double> B;
	static Array<double> X;
	double C;

	error_sum = 0;
	error_max = 0;

	// sum errors for all points in the domain. Lots of little systems of equations.
	X.resize(single_nfuncs);
	A.resize(X.s,X.s);
	B.resize(X.s);

	for (int i = 0; i < X.s; i++)
	{
		for (int j = 0; j < X.s; j++)
			A(i,j) = bb(seti[i],seti[j]);
	}

	for (int k = 0; k < patch.samples.data_size; k++)
	{
		Sample &sample = patch.samples.data[k];

		// build A, B, C
		C = sample.hh;
		for (int i = 0; i < X.s; i++)
			B[i] = sample.hb[seti[i]];

		// build X
#if 1
		for (int i = 0; i < X.s; i++)
		{
			X[i] = 0;
			for (int ii = 0; ii < single_npoly; ii++)
				X[i] += X_in[i*single_npoly + ii] * poly_func[ii](sample.t);
		}
#else
		//int dposv_joe(char uplo, integer n, integer nrhs, doublereal *a, integer lda, doublereal *b, integer ldb)
		for (int i = 0; i < X.s; i++)
			X[i] = B[i];
		int info = dposv_joe('L', B.s, 1, A.data, B.s, X.data, B.s);
#endif

		// eval error
		double error = C;
		for (int i = 0; i < X.s; i++)
		{
			double Ax = 0;
			for (int j = 0; j < X.s; j++)
				Ax += X[j] * A(i,j);
			error += X[i] * (Ax - 2*B[i]);
		}

		error_sum += error;
		error_max = max(error_max, error);
	}

	error_sum /= patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);
}


Array2D<double> A;
Array<double> B;
double C;
SetError sete;


int get_sets_tree(SetError &best_set,
				   BoundedArray<int, single_nfuncs> &seti, 
				   Array<BasisParam> &basis,
				   Array<Affinity> &order,
				   int mini,
				   int nleaves,
				   Patch &patch)
{
	if (nleaves >= single_maxleaves)
		return nleaves;

	if (seti.s == single_nfuncs)
	{
		// resize
		BoundedArray<double, single_nfuncs*single_npoly> &X = sete.coeffs;
		X.resize(single_nfuncs * single_npoly);

		//** build matrices
		// add up A
		Array<double> &Bc = Bpart[seti[0]];
		Array2D<double> &Acc = Apart(seti[0], seti[0]);

		for (int r = 0; r < single_nfuncs-1; r++)
		{
			Array2D<double> &Auc = Apart(seti[r+1], seti[0]);
			for (int i = 0; i < single_nfuncs-1; i++)
			{
				Array2D<double> &Auu = Apart(seti[r+1], seti[i+1]);
				Array2D<double> &Acu = Apart(seti[0], seti[i+1]);

				for (int s = 0; s < single_npoly; s++)
				{
					int rs = r*single_npoly + s;
					for (int j = 0; j < single_npoly; j++)
					{
						int ij = i*single_npoly + j;
						A(rs,ij) = Auu(s, j) - Auc(s,j) - Acu(s,j) + Acc(s,j);
					}
				}
			}
		}

		//** add up B
		for (int i = 0; i < single_nfuncs-1; i++)
		{
			Array<double> &Bu = Bpart[seti[i+1]];
			Array2D<double> &Auc = Apart(seti[i+1], seti[0]);
			for (int j = 0; j < single_npoly; j++)
			{
				int ij = i*single_npoly + j;
				B[ij] = Bu[j] - Auc(j, 0) + Acc(j, 0) - Bc[j];
			}
		}

		//** add up C
		C = Cpart + Acc(0,0) - 2*Bc[0];

		//** solve for values
		for (int i = 0; i < B.s; i++)
			X[i+single_npoly] = B[i];

		int info = dposv_joe('L', B.s, 1, A.data, B.s, X.data+single_npoly, B.s);

		if (info) // failed to solve, so there is a linear dependence. SKIP: BAD BASIS SET
			return nleaves;
			
		// construct constrained values
		X[0] = 1;
		for (int j = 1; j < single_npoly; j++)
			X[j] = 0;

		for (int i = 1; i < single_nfuncs; i++)
		{
			for (int j = 0; j < single_npoly; j++)
				X[j] -= X[i*single_npoly + j];
		}

		// calculate error
		sete.error = C;
		for (int i = 0; i < B.s; i++)
			sete.error -= B[i]*X[i+single_npoly];

		// multiply by differential area of sample
		sete.error /= patch.samples.data_size * (4*npatches_xy*npatches_xy*npatches_z);

		//** store
		if (nleaves % 10000 == 0)
			printf(" nleaves = %d\r", nleaves);

		if (sete.error < best_set.error)
		{
			printf("new best error %f at leaf %d                 \n", sete.error, nleaves);

			best_set.seti = seti;
			best_set.error = sete.error;
			best_set.coeffs = sete.coeffs;
			best_set.index = nleaves;

			print_set(best_set, basis);
			
			double error_sum = 0;
			double error_max = 0;
			calc_error_max(error_sum, error_max, patch, seti, X);
			best_set.error_max = error_max;
			printf("checking again (error_sum = %f, error_max = %f)                 \n", error_sum, error_max);
		}

		nleaves++;
		return nleaves;
	}

	for (int i = basis.s-1; i >= mini; i--)
	{
		seti.push_back(order[i].index);
		nleaves = get_sets_tree(best_set, seti, basis, order, i+1, nleaves, patch);
		seti.pop_back();
	}

	return nleaves;
}

void get_sets(Patch &patch, Array<BasisParam> &basis)
{
	patch.best_set.error = 1e30;

	A.resize((single_nfuncs-1) * single_npoly, (single_nfuncs-1) * single_npoly);
	B.resize((single_nfuncs-1) * single_npoly);

	BoundedArray<int, single_nfuncs> seti;
	get_sets_tree(patch.best_set, seti, basis, patch.affinity, 0, 0, patch);
	
	// save basis indices
	patch.best_set.basis_idx.s = 0;
	for (int i = 0; i < single_nfuncs; i++)
		patch.best_set.basis_idx.push_back(basis[patch.best_set.seti[i]]);
}

void write_expanded(Array<SetError> &best_sets)
{
	assert(single_npoly == 4);
	assert(npatches_xy == 2);

	FILE *find = fopen("results_expanded_indices.txt", "wb");
	FILE *fcoeff = fopen("results_expanded_coeffs.txt", "wb");
	printf("opening expanded files for writing\n");

	fprintf(find, "int indices[%d*%d*%d*%d][3] = {", npatches_xy*2, npatches_xy*2, npatches_z, single_nfuncs);
	fprintf(fcoeff, "double coeffs[%d*%d*%d*%d][4] = {", npatches_xy*2, npatches_xy*2, npatches_z, single_nfuncs);
	for (int pzi = 0; pzi < npatches_z; pzi++)
	{
		for (int pyi = 0; pyi < npatches_xy*2; pyi++)
		{
			for (int pxi = 0; pxi < npatches_xy*2; pxi++)
			{
				int px = pxi;
				int py = pyi;
				int pz = pzi;

				// find symmetry
				bool flipx = false, flipy = false, flipxy = false;

				if (px >= 2)
				{
					flipx = true;
					px = 3 - px;
				}
				if (py >= 2)
				{
					flipy = true;
					py = 3 - py;
				}
				if (py > px)
				{
					flipxy = true;
					swap(px, py);
				}

				int sind = px + py + 3*pz;

				// write patch
				SetError &s = best_sets[sind];

				for (int i = 0; i < single_nfuncs; i++)
				{
					if (i > 0 || pxi > 0 || pyi > 0 || pzi > 0)
					{
						fprintf(find, ",");
						fprintf(fcoeff, ",");
					}

					int ox = s.basis_idx[i].i;
					int oy = s.basis_idx[i].j;
					int oz = s.basis_idx[i].k;

					double c = s.coeffs[i*single_npoly + 0];
					double lx = s.coeffs[i*single_npoly + 1];
					double ly = s.coeffs[i*single_npoly + 2];
					double lz = s.coeffs[i*single_npoly + 3];

					if (flipxy)
					{
						swap(ox, oy);
						swap(lx, ly);
					}

					if (flipx)
					{
						ox = (1<<oz) - 1 - ox;
						c += lx;
						lx = -lx;
					}
					if (flipy)
					{
						oy = (1<<oz) - 1 - oy;
						c += ly;
						ly = -ly;
					}

					fprintf(find, "{%d,%d,%d}", ox, oy, oz);

					fprintf(fcoeff, "{%16.15f", c);
					fprintf(fcoeff, ",%16.15f,%16.15f,%16.15f}", lx, ly, lz);
				}
			}
		}
	}
	fprintf(find, "};\n\n");
	fprintf(fcoeff, "};\n\n");

	fclose(find);
	fclose(fcoeff);
}

}

using namespace Single;

void calc_single_fetches()
{
	//========== choose basis functions
	Array<BasisParam> basis;
	choose_basis_functions(basis);

	// select which patches to evaluate. -1 means all patches in that dimension.
	int inpatch_i = -1, inpatch_j = -1, inpatch_k = -1;
	//int inpatch_i = 0, inpatch_j = 0, inpatch_k = 0;

	Array<SetError> best_sets;
	best_sets.reserve(6);

	char outname[256];

	// build full linear systems
	for (int pk = 0; pk < npatches_z; pk++)
	{
		if (inpatch_k != -1 && inpatch_k != pk)
			continue;

		for (int pj = 0; pj < npatches_xy; pj++)
		{
			if (inpatch_j != -1 && inpatch_j != pj)
				continue;

			for (int pi = 0; pi < npatches_xy; pi++)
			{
				if (inpatch_i != -1 && inpatch_i != pi)
					continue;

				if (pj > pi)
					continue;
				
				sprintf(outname, "results%s_patch_%d_%d_%d.txt", max_mip_level == 1 ? "_low" : "", pi, pj, pk);
				fr = fopen(outname, "wb");
				printf("opening file %s for writing\n", outname);

				printf("starting now for patch %d %d %d\n", pi, pj, pk);
				fprintf(fr, "starting now for patch %d %d %d\n", pi, pj, pk);
				double t0 = get_time();


				// patch offsets and scales
				double poi = pi / (double)(2 * npatches_xy);
				double psi = 1 / (double)(2 * npatches_xy);

				double poj = pj / (double)(2 * npatches_xy);
				double psj = 1 / (double)(2 * npatches_xy);

				//double pok = 1 + pk / (double)(npatches_z);
				double pok = pk / (double)(npatches_z);
				double psk = 1 / (double)(npatches_z);
				
				//========== build overlap matrices
				Patch patch;

				printf("size = %d\n", patch.affinity.s);
				build_overlaps(patch.samples, bb, basis, poi, psi, poj, psj, pok, psk, nsamples);

				double t1 = get_time();
				printf("build overlaps = %f\n", t1 - t0);
				fprintf(fr, "build overlaps = %f\n", t1 - t0);

				//========== build parts of the system of equations
				build_parts(patch, basis);

				double t2 = get_time();
				printf("build parts = %f\n", t2 - t1);
				fprintf(fr, "build parts = %f\n", t2 - t1);
				
				//========== sort by affinity heuristic
				patch.affinity.resize(basis.s);
				for (int bi = 0; bi < basis.s; bi++)
				{
					patch.affinity[bi].index = bi;
					patch.affinity[bi].affinity = 0;
				}

				// solve for fitting errors of individual basis functions
				Array<double> X;
				A.resize(single_npoly, single_npoly);
				B.resize(single_npoly);
				X.resize(single_npoly);

				for (int bi = 0; bi < basis.s; bi++)
				{
					// SOLVE LINEAR SYSTEM FOR POLYNOMIAL FIT
					A = Apart(bi,bi);
					B = Bpart[bi];
					X = B;

					int info = dposv_joe('L', B.s, 1, A.data, B.s, X.data, B.s);
					assert(!info);

					// calculate error
					double error = Cpart;
					for (int i = 0; i < B.s; i++)
						error -= B[i]*X[i];

					patch.affinity[bi].affinity = -error;
				}

				sort(patch.affinity.data, patch.affinity.data + patch.affinity.s);

				//========== get sets with errors
				get_sets(patch, basis);

				double t3 = get_time();
				printf("find sets = %f\n", t3 - t2);
				fprintf(fr, "find sets = %f\n", t3 - t2);

				printf("total = %f\n", t3 - t0);
				fprintf(fr, "total = %f\n", t3 - t0);
				
				// store best set
				patch.best_set.subdomain[0] = pi;
				patch.best_set.subdomain[1] = pj;
				patch.best_set.subdomain[2] = pk;

				best_sets.push_back(patch.best_set);


				fclose(fr);
			}
		}
	}

	// print total error
	printf("nfuncs = %d, npoly = %d, npatches_xy = %d (%d), npatches_z = %d\n", single_nfuncs, single_npoly, npatches_xy, npatches_xy*2, npatches_z);
	double total_error = 0;
	double maximal_error = 0;
	for (int i = 0; i < best_sets.s; i++)
	{
		if (best_sets[i].subdomain[0] == best_sets[i].subdomain[1])
			total_error += 4*best_sets[i].error; // only 4x on diagonal, because we cover the whole square rather than just a triangle
		else
			total_error += 8*best_sets[i].error;

		maximal_error = max(maximal_error, best_sets[i].error_max);
	}
	printf("The total error over all patches is %f\n", total_error);
	printf("The maximal error over all patches is %f\n", maximal_error);

	//** output C++ arrays of best sets in symmetric form 6
	sprintf(outname, "results%s_all.txt", max_mip_level == 1 ? "_low" : "");
	fr = fopen(outname, "wb");
	printf("opening file %s for writing\n", outname);
	fprintf(fr, "The total error over all patches is %f\n\n", total_error);
	fprintf(fr, "The maximal error over all patches is %f\n", maximal_error);

	// write indices
	fprintf(fr, "static int indices[%d][%d][3] = {\n", best_sets.s, single_nfuncs);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetError &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < single_nfuncs; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			fprintf(fr, "{%d,%d,%d}", s.basis_idx[i].i, s.basis_idx[i].j, s.basis_idx[i].k);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	// write coefficients
	fprintf(fr, "static double coeffs[%d][%d] = {\n", best_sets.s, single_nfuncs*single_npoly);
	for (int pi = 0; pi < best_sets.s; pi++)
	{
		SetError &s = best_sets[pi];
		
		fprintf(fr, "\t{");

		for (int i = 0; i < s.coeffs.s; i++)
		{
			if (i > 0)
				fprintf(fr, ",");
			fprintf(fr, "%16.15f", s.coeffs[i]);
		}
		
		if (pi < best_sets.s - 1)
			fprintf(fr, "},\n");
		else
			fprintf(fr, "}\n");
	}
	fprintf(fr, "};\n\n");

	fclose(fr);

	
	//** output non-symmetric arrays, expanded
	//write_expanded(best_sets);
}
