#include <stdio.h>
#include <math.h>

#include "optimize.h"

void get_soln_bi(Array<BasisParam> &basis, Array<double> &X, double *t)
{
	// find indices of the correct functions
	const int ind00 = 0, ind10 = 1, ind01 = 2, ind11 = 3; // indXYZ

	// top level is same for all values of t we use
	X.resize(4);
	basis.resize(4);

	if (t[2] < .5)
	{
		basis[ind00].i = -1;
		basis[ind00].j = -1;
		basis[ind00].k = 0;

		basis[ind10].i = 0;
		basis[ind10].j = -1;
		basis[ind10].k = 0;

		basis[ind01].i = -1;
		basis[ind01].j = 0;
		basis[ind01].k = 0;

		basis[ind11].i = 0;
		basis[ind11].j = 0;
		basis[ind11].k = 0;

		double tx = t[0] + .5;
		double ty = t[1] + .5;

		X[0] = (1-tx)*(1-ty);
		X[1] = tx*(1-ty);
		X[2] = (1-tx)*ty;
		X[3] = tx*ty;
	}
	else
	{
		// the bottom parts have four different sets of coefficents
		double bx = 2*t[0] + .5;
		double by = 2*t[1] + .5;

		if (t[0] <= .25 && t[1] <= .25)
		{
			basis[ind00].i = -1;
			basis[ind00].j = -1;
			basis[ind00].k = 1;

			basis[ind10].i = 0;
			basis[ind10].j = -1;
			basis[ind10].k = 1;

			basis[ind01].i = -1;
			basis[ind01].j = 0;
			basis[ind01].k = 1;

			basis[ind11].i = 0;
			basis[ind11].j = 0;
			basis[ind11].k = 1;
		}
		else if (t[0] > .25 && t[1] <= .25)
		{
			bx -= 1;
			basis[ind00].i = 0;
			basis[ind00].j = -1;
			basis[ind00].k = 1;

			basis[ind10].i = 1;
			basis[ind10].j = -1;
			basis[ind10].k = 1;

			basis[ind01].i = 0;
			basis[ind01].j = 0;
			basis[ind01].k = 1;

			basis[ind11].i = 1;
			basis[ind11].j = 0;
			basis[ind11].k = 1;
		}
		else if (t[0] <= .25 && t[1] > .25)
		{
			by -= 1;
			basis[ind00].i = -1;
			basis[ind00].j = 0;
			basis[ind00].k = 1;

			basis[ind10].i = 0;
			basis[ind10].j = 0;
			basis[ind10].k = 1;

			basis[ind01].i = -1;
			basis[ind01].j = 1;
			basis[ind01].k = 1;

			basis[ind11].i = 0;
			basis[ind11].j = 1;
			basis[ind11].k = 1;
		}
		else if (t[0] > .25 && t[1] > .25)
		{
			bx -= 1;
			by -= 1;
			basis[ind00].i = 0;
			basis[ind00].j = 0;
			basis[ind00].k = 1;

			basis[ind10].i = 1;
			basis[ind10].j = 0;
			basis[ind10].k = 1;

			basis[ind01].i = 0;
			basis[ind01].j = 1;
			basis[ind01].k = 1;

			basis[ind11].i = 1;
			basis[ind11].j = 1;
			basis[ind11].k = 1;
		}

		X[0] = (1-bx)*(1-by);
		X[1] = bx*(1-by);
		X[2] = (1-bx)*by;
		X[3] = bx*by;
	}
}

void get_soln_tri(Array<BasisParam> &basis, Array<double> &X, double *t)
{
	// find indices of the correct functions
	const int ind000 = 0, ind100 = 1, ind010 = 2, ind110 = 3, ind001 = 4, ind101 = 5, ind011 = 6, ind111 = 7; // indXYZ

	// top level is same for all values of t we use
	X.resize(8);
	basis.resize(8);

	basis[ind000].i = -1;
	basis[ind000].j = -1;
	basis[ind000].k = 0;

	basis[ind100].i = 0;
	basis[ind100].j = -1;
	basis[ind100].k = 0;

	basis[ind010].i = -1;
	basis[ind010].j = 0;
	basis[ind010].k = 0;

	basis[ind110].i = 0;
	basis[ind110].j = 0;
	basis[ind110].k = 0;


	double tx = t[0] + .5;
	double ty = t[1] + .5;

	X[0] = (1-tx)*(1-ty)*(1-t[2]);
	X[1] = tx*(1-ty)*(1-t[2]);
	X[2] = (1-tx)*ty*(1-t[2]);
	X[3] = tx*ty*(1-t[2]);

	// the bottom parts have four different sets of coefficents
	double bx = 2*t[0] + .5;
	double by = 2*t[1] + .5;

	if (t[0] <= .25 && t[1] <= .25)
	{
		basis[ind001].i = -1;
		basis[ind001].j = -1;
		basis[ind001].k = 1;

		basis[ind101].i = 0;
		basis[ind101].j = -1;
		basis[ind101].k = 1;

		basis[ind011].i = -1;
		basis[ind011].j = 0;
		basis[ind011].k = 1;

		basis[ind111].i = 0;
		basis[ind111].j = 0;
		basis[ind111].k = 1;
	}
	else if (t[0] > .25 && t[1] <= .25)
	{
		bx -= 1;
		basis[ind001].i = 0;
		basis[ind001].j = -1;
		basis[ind001].k = 1;

		basis[ind101].i = 1;
		basis[ind101].j = -1;
		basis[ind101].k = 1;

		basis[ind011].i = 0;
		basis[ind011].j = 0;
		basis[ind011].k = 1;

		basis[ind111].i = 1;
		basis[ind111].j = 0;
		basis[ind111].k = 1;
	}
	else if (t[0] <= .25 && t[1] > .25)
	{
		by -= 1;
		basis[ind001].i = -1;
		basis[ind001].j = 0;
		basis[ind001].k = 1;

		basis[ind101].i = 0;
		basis[ind101].j = 0;
		basis[ind101].k = 1;

		basis[ind011].i = -1;
		basis[ind011].j = 1;
		basis[ind011].k = 1;

		basis[ind111].i = 0;
		basis[ind111].j = 1;
		basis[ind111].k = 1;
	}
	else if (t[0] > .25 && t[1] > .25)
	{
		bx -= 1;
		by -= 1;
		basis[ind001].i = 0;
		basis[ind001].j = 0;
		basis[ind001].k = 1;

		basis[ind101].i = 1;
		basis[ind101].j = 0;
		basis[ind101].k = 1;

		basis[ind011].i = 0;
		basis[ind011].j = 1;
		basis[ind011].k = 1;

		basis[ind111].i = 1;
		basis[ind111].j = 1;
		basis[ind111].k = 1;
	}

	X[4] = (1-bx)*(1-by)*t[2];
	X[5] = bx*(1-by)*t[2];
	X[6] = (1-bx)*by*t[2];
	X[7] = bx*by*t[2];
}

struct IntegrationDomain
{
	double start[3];
	double end[3];
	int nsamples;
};

void calc_error( void (*get_soln)(Array<BasisParam> &basis, Array<double> &X, double *t), IntegrationDomain domain, double &error_sum, double &error_max)
{
	Array<BasisParam> basis;
	Array<double> X;

	Array2D<double> A;
	Array<double> B;
	double C;

	double t[3], h[3], a[3], b[3];

	double step[3];
	step[0] = (domain.end[0] - domain.start[0]) / domain.nsamples;
	step[1] = (domain.end[1] - domain.start[1]) / domain.nsamples;
	step[2] = (domain.end[2] - domain.start[2]) / domain.nsamples;

	// Trilinear error
	error_sum = 0;
	error_max = 0;
	for (t[0] = domain.start[0] + step[0]*.5; t[0] < domain.end[0]; t[0] += step[0])
	{
		for (t[1] = domain.start[1] + step[1]*.5; t[1] < domain.end[1]; t[1] += step[1])
		{
			for (t[2] = domain.start[2] + step[2]*.5; t[2] < domain.end[2]; t[2] += step[2])
			{
				get_soln(basis, X, t);

				// evaluate error
				A.resize(X.s,X.s);
				B.resize(X.s);

				h[0] = t[0];
				h[1] = t[1];
				//h[2] = 1 + t[2];
				h[2] = pow(2.0, t[2]);

				C = filter_squared(h) * (h[2]*h[2]*h[2]*h[2]);
				for (int i = 0; i < X.s; i++)
				{
					a[0] = basis[i].i;
					a[1] = basis[i].j;
					a[2] = basis[i].k;

					B[i] = basis_times_filter(a,h) * (pow(4.0, basis[i].k) * h[2]*h[2]);

					for (int j = 0; j < X.s; j++)
					{
						b[0] = basis[j].i;
						b[1] = basis[j].j;
						b[2] = basis[j].k;

						A(i,j) = basis_times_basis(a,b) * (pow(4.0, basis[i].k) * pow(4.0, basis[j].k));
					}
				}

				double error = C;
				for (int i = 0; i < X.s; i++)
				{
					double Ax = 0;
					for (int j = 0; j < X.s; j++)
						Ax += X[j] * A(i,j);
					error += X[i] * (Ax - 2*B[i]);
				}

				error_sum += error;
				error_max = max(error_max, error);
			}
		}
	}

	error_sum *= step[0]*step[1]*step[2];
}

void calc_bilinear()
{
	IntegrationDomain domain;
	domain.start[0] = domain.start[1] = domain.start[2] = 0;
	domain.end[0] = domain.end[1] = .5;
	domain.end[2] = 1;
	domain.nsamples = 10;
	double error_sum, error_max;
	calc_error(get_soln_bi, domain, error_sum, error_max);
	printf("bilinear error total = %f\n", 4*error_sum);
	printf("bilinear error maximal = %f\n", error_max);
}

void calc_trilinear()
{
	IntegrationDomain domain;
	domain.start[0] = domain.start[1] = domain.start[2] = 0;
	domain.end[0] = domain.end[1] = .5;
	domain.end[2] = 1;
	domain.nsamples = 10;
	double error_sum, error_max;
	calc_error(get_soln_tri, domain, error_sum, error_max);
	printf("trilinear error total = %f\n", 4*error_sum);
	printf("trilinear error maximal = %f\n", error_max);
}

