#include "complete_squares.h"
#include <cmath>
#include <vector>

using namespace std;


#if 1
void drop_into(Array<double> &r, Array2D<double> &a, int a_size, int k)
{
	r.resize(a_size-1);

	for (int i = 0; i < k; i++)
		r[i] = a(i,k);
	
	for (int i = k+1; i < a_size; i++)
		r[i-1] = a(i,k);
}

template<class T>
void drop(Array<T> &a, int k)
{
	for (int i = k+1; i < a.s; i++)
		a[i-1] = a[i];
	a.pop_back();
}

void drop(Array2D<double> &a, int k)
{
	assert(a.s[0] == a.s[1]);
	int s = a.s[0];
	for (int i = k+1; i < s; i++)
	{
		for (int j = 0; j < k; j++)
		{
			a(i-1,j) = a(i,j);
			a(j,i-1) = a(i,j);
		}
	}
	for (int i = k+1; i < s; i++)
	{
		for (int j = k+1; j < s; j++)
			a(i-1,j-1) = a(i,j);
	}
}

double complete_squares(Array<CompletedSquare> &squares, Array2D<double> &Ain, Array<double> &Bin, double C)
{
	assert(Ain.s[0] == Ain.s[1] && Ain.s[0] == Bin.s);
	
	static Array2D<double> A;
	static Array<double> B;
	static Array<double> pbl;
	static Array<int> IND;

	A = Ain;
	B = Bin;

	squares.reserve(A.s[0]);

	IND.resize(A.s[0]);
	for (int i = 0; i < IND.s; i++)
		IND[i] = i;

	for (int row = A.s[0]-1; row >= 0; row--)
	{
		// get pivot
		int pivot = 0;
		//vector<double> A_diag;
		for (int i = 0; i < row+1; i++)
		{
			double ap = A(pivot,pivot);
			double ai = A(i,i);
			//A_diag.push_back(ap);
			if (ai > ap)
				pivot = i;
		}

		// get a and b parts
		double pa = A(pivot, pivot);
		if (pa < 1e-6)
			return C;
		double spa = sqrt(pa);
		double ipa = 1.0 / pa;
		double ispa = 1.0 / spa;

		drop_into(pbl, A, row+1, pivot);

		double pbc = B[pivot];

		// calculate completed square
		CompletedSquare &r = squares.push_back();
		r.linear.resize(A.s[0]);
		for (int i = 0; i < A.s[0]; i++)
			r.linear[i] = 0;
		for (int i = 0; i < pivot; i++)
			r.linear[IND[i]] = pbl[i] * ispa;
		r.linear[IND[pivot]] = spa;
		for (int i = pivot; i < row; i++)
			r.linear[IND[i+1]] = pbl[i] * ispa;
		
		r.constant = pbc * .5 * ispa;

		// update remaining constant part of equation by subtracting b^2/4a
		drop(B, pivot);
		drop(A, pivot);
		drop(IND, pivot);

		C -= pbc*pbc*.25*ipa;

		for (int i = 0; i < row; i++)
		{
			B[i] -= pbl[i]*pbc*ipa;

			for (int j = 0; j < row; j++)
				A(j,i) -= pbl[i]*pbl[j]*ipa;
		}
	}

	return C;
}
#else
void complete_squares(Array<CompletedSquare> &squares, Array2D<double> &Ain, Array<double> &Bin)
{
	assert(Ain.s[0] == Ain.s[1] && Ain.s[0] == Bin.s);
	
	static Array2D<double> A;
	static Array<double> B = Bin;

	A = Ain;
	B = Bin;

	squares.resize(A.s[0]);

	for (int row = A.s[0]-1; row >= 0; row--)
	{
		// get a and b parts
		double pa = A(row, row);
		double spa = sqrt(pa);
		double ispa = 1.0 / spa;
		double ipa = 1.0 / pa;

		double *pbl = &A(0, row);

		double pbc = B[row];

		// calculate completed square
		CompletedSquare &r = squares[row];
		r.linear.resize(row+1);
		for (int i = 0; i < row; i++)
			r.linear[i] = pbl[i] * ispa;
		r.linear[row] = spa;

		r.constant = pbc * .5 * ispa;

		// update remaining constant part of equation by subtracting b^2/4a
		for (int i = 0; i < row; i++)
		{
			B[i] -= pbl[i]*pbc*ipa;

			for (int j = 0; j < row; j++)
				A(j,i) -= pbl[i]*pbl[j]*ipa;
		}
	}
}
#endif
