#pragma once

template <class T>
double integrate_trapezoid(T func, double a, double b, int N)
{
	double h = (b - a) / N;
	double sum = 0;

	double p1, f0, f1;

	p1 = a;
	f1 = func(p1);

	for (int i = 1; i <= N; i++)
	{
		p1 += h;

		f0 = f1;
		f1 = func(p1);

		sum += f0 + f1;
	}

	return sum * h / 2.0;
}

template <class T>
double integrate_trapezoid2(T func, double a, double b, int N)
{
	// this simplified inner loop from above is inexplicably slower. 
	// maybe it will be better on a different CPU
	double h = (b - a) / N;
	double p1, f0, f1;

	p1 = a;
	f1 = func(p1);

	double sum = .5 * f1;

	for (int i = 1; i < N; i++)
	{
		p1 += h;
		f1 = func(p1);

		sum += f1;
	}

	p1 += h;
	f1 = func(p1);

	sum += .5 * f1;

	return sum * h;
}

template <class T>
double integrate_simpson(T func, double a, double b, int N)
{
	double h = (b - a) / N;
	double h2 = .5 * h;
	double h6 = h / 6.0;
	double sum = 0;

	double p0, p1, p2, f0, f1, f2;

	p2 = a;
	f2 = func(p2);

	for (int i = 1; i <= N; i++)
	{
		p1 = p2 + h2;
		p2 += h;

		f0 = f2;
		f1 = func(p1);
		f2 = func(p2);

		sum += h6 * (f0 + 4*f1 + f2);
	}

	return sum;
}

template <class T>
double integrate2_box(T func, double ax, double bx, double ay, double by, int M, int N)
{
	double hx = (bx - ax) / M;
	double hx2 = .5 * hx;
	
	double hy = (by - ay) / M;
	double hy2 = .5 * hy;

	double sum = 0;

	double py = ay + hy2;
	for (int j = 0; j < N; j++)
	{
		double px = ax + hx2;
		for (int i = 0; i < M; i++)
		{
			double v = func(px, py);
			sum += v;
			px += hx;
		}

		py += hy;
	}

	return sum*hx*hy;
}