#pragma once

#include <string>

#include "filterintegrals.h"
#include "boundedarray.h"
#include "array.h"
#include "array2d.h"
#include "array3d.h"
#include "timer.h"
#include "complete_squares.h"

using namespace std;

//=========== top-level functions
void calc_bilinear();
void calc_trilinear();
void calc_single_fetches();
void calc_bilinear_fetches();
void calc_bilinear_hw_fetches();

//=========== structures
struct BasisParam
{
	int i, j, k;

	bool operator==(BasisParam a)
	{
		return i == a.i && j == a.j && k == a.k;
	}
};

struct BasisIndexBi
{
	int idx[4]; // indexes basisparam
};

struct Affinity
{
	double affinity;
	int index;

	bool operator<(const Affinity &a) const
	{
		return affinity < a.affinity;
	}
};

struct Sample
{
	double t[3]; // location within patch [0,1]^3
	double h[3]; // parameters of function
	double hh;
	Array<double> hb;
};

struct GridBasisIndices
{
	int minidx;
	int maxidx;
	Array2D<int> grid;
};

extern GridBasisIndices grid_basis_indices[5]; // indexed by mip level

//=========== partitioning of the domain
const int nsamples = 5;
const int npatches_xy = 4 /2; // *2 // This value is for the number of patches in the quadrant, so multiply by 2 to get total number on a side
const int npatches_z = 2;

const int max_mip_level = 2; // normally is 2 (for top), but can be 1 (for bottom)

const int bilinear_npoly = 4; // 1, 4, 10
const int bilinear_nfuncs = 2;
const int bilinear_maxleaves = 500*1000;
const int bilinear_iter_max = 1000;

//=========== helper functions
void choose_basis_functions(Array<BasisParam> &basisparams);
void choose_hw_basis_functions(Array<BasisParam> &basisparams);
void choose_bilinear_basis_indices(Array<BasisIndexBi> &basisindexbi, Array<BasisParam> &basisparams);
void build_overlaps(Array3D<Sample> &samples, Array2D<double> &bb, Array<BasisParam> &basisparams, double poi, double psi, double poj, double psj, double pok, double psk, int ns);

void vec_print(double *A, int s, FILE *fa);
void vec_print(double *A, int s, string fn);
void vec_print(Array<double> &A, string fn);
void mat_print(Array2D<double> &A, string fn);

double frand0();
